/**
 * Breakthrough - A GUI interface for playing the game Breakthrough.  
 * Depends on and is useful for testing class <code>BreakthroughState</code>.
 * Red/black pieces correspond to the 1st/2nd white/black player, respectively.
 *
 * @author Todd Neller
 * @version 1.0

Copyright (C) 2014 Todd Neller

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

Information about the GNU General Public License is available online at:
  http://www.gnu.org/licenses/
To receive a copy of the GNU General Public License, write to the Free
Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.

 */

import java.awt.*;
import java.awt.event.*;
import java.util.Stack;

import javax.swing.*;

public class Breakthrough extends JFrame {
	private static final long serialVersionUID = -8595697962612897813L;

	public Breakthrough() {
		super();
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setTitle("Breakthrough");
		add(new BreakthroughPanel(), BorderLayout.CENTER);
		pack();
		setVisible(true);
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		new Breakthrough();
	}
}

class BreakthroughPanel extends JPanel {
	private static final long serialVersionUID = 7541199107576903753L;
	private static final boolean flipBoard = true, reportMoves = true, undoEnabled = true;
	boolean isMovingPiece = false; // whether or not piece is currently being moved (i.e. dragged)
	int pressRow, pressCol; // row, col of mouse press
	int mouseX, mouseY; // current position of dragging mouse
	static final int SQUARE1 = 0, SQUARE2 = 1, PLAYER1 = 2, PLAYER2 = 3, IMAGE_SIZE = 80;
	static final String[] IMAGE_FILENAMES = {"whiteMarble80.gif", "greenMarble80.gif", "red80.gif", "black80.gif"};
	Image[] images = new Image[4]; 
	BreakthroughState state = new BreakthroughState();
	Stack<BreakthroughState> stateStack = new Stack<BreakthroughState>();

	BreakthroughPanel() {
		// initialize panel
		setBackground(Color.white);
		setPreferredSize(new Dimension(640, 640));
		setSize(640, 640);
		addMouseListener(new MouseHandler());
		addMouseMotionListener(new MouseMotionHandler());
		if (undoEnabled)
			addKeyListener(new UndoHandler());
		requestFocusInWindow();

		// load images
		for (int i = 0; i < IMAGE_FILENAMES.length; i++)
			images[i] = getToolkit().getImage(IMAGE_FILENAMES[i]);
		MediaTracker mt = new MediaTracker(this);
		for (int i = 0; i < IMAGE_FILENAMES.length; i++)
			mt.addImage(images[i], i);
		try {
			mt.waitForAll();
		} catch (Exception e) {
			System.err.println("Exception while loading image.");
		}
		// check that images loaded
		for (int i = 0; i < IMAGE_FILENAMES.length; i++)
			if (images[i].getWidth(this) == 0) {
				System.err.printf("Image \"%s\" not loaded.\n", IMAGE_FILENAMES[i]);
				System.exit(0);
			}
	}

	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D) g;
		// draw each board square
		for (int row = 0; row < BreakthroughState.SIZE; row++)
			for (int col = 0; col < BreakthroughState.SIZE; col++) {
				int y = IMAGE_SIZE * (flipBoard ? BreakthroughState.SIZE - 1 - row : row);
				int x = IMAGE_SIZE * col;
				g2.drawImage(images[((row + col) % 2 == 0) ? SQUARE1 : SQUARE2], x, y, this); // draw background checkerboard square
				int contents = state.getPiece(row, col);
				// draw square contents (if any)

				if (contents == BreakthroughState.WHITE && (!isMovingPiece || pressRow != row || pressCol != col)) // if white piece that's not in motion
					g2.drawImage(images[PLAYER1], x, y, this); // draw white piece
				else if (contents == BreakthroughState.BLACK && (!isMovingPiece || pressRow != row || pressCol != col)) // if black piece that's not in motion
					g2.drawImage(images[PLAYER2], x, y, this); // draw black piece
			}
		if (isMovingPiece) // if piece is moving
			g2.drawImage(images[state.getPiece(pressRow, pressCol) == BreakthroughState.WHITE ? PLAYER1 : PLAYER2], 
					mouseX - IMAGE_SIZE / 2, mouseY - IMAGE_SIZE / 2, this); // draw piece
	}

	// a mouse handler to trigger mouse press/release events
	class MouseHandler extends MouseAdapter {

		@Override
		public void mousePressed(MouseEvent event) {
			requestFocusInWindow();
			pressRow = event.getY() / BreakthroughPanel.IMAGE_SIZE; // discretize source board row, col
			pressCol = event.getX() / BreakthroughPanel.IMAGE_SIZE;
			if (flipBoard)
				pressRow = BreakthroughState.SIZE - pressRow - 1;
			if (state.getPiece(pressRow, pressCol) == BreakthroughState.WHITE || state.getPiece(pressRow, pressCol) == BreakthroughState.BLACK)
				isMovingPiece = true; // change state to reflect piece motion potential
		}

		@Override
		public void mouseReleased(MouseEvent event) {
			requestFocusInWindow();
			isMovingPiece = false; // change state to reflect piece motion impossibility
			int releaseRow = event.getY() / BreakthroughPanel.IMAGE_SIZE; // discretize destination row, col
			int releaseCol = event.getX() / BreakthroughPanel.IMAGE_SIZE;
			if (flipBoard)
				releaseRow = BreakthroughState.SIZE - releaseRow - 1;
			// move
			if (reportMoves)
//				//System.out.printf("%d %d %d %d\n", pressRow, pressCol, releaseRow, releaseCol);
				System.out.println((char) ('a' + pressCol) + Integer.toString(pressRow + 1) + "-" + (char) ('a' + releaseCol) + Integer.toString(releaseRow + 1));
			BreakthroughState copy = (BreakthroughState) state.clone();
			if (state.makeMove(pressRow, pressCol, releaseRow, releaseCol)) { // attempt move, checking legality
				stateStack.push(copy); // push copy for undo
			}	
			repaint(); // repaint
			if (state.getWinner() != BreakthroughState.NONE) {
				JOptionPane.showMessageDialog(null, String.format("%s wins!\n", state.getWinner() == BreakthroughState.BLACK ? "Black" : "Red"));
			}
		}
	}


		// a mouse motion handler to track piece movement and force repainting during movement
		// based on example from http://www.java2s.com/Code/Java/2D-Graphics-GUI/Imagewithmousedragandmoveevent.htm
		class MouseMotionHandler extends MouseMotionAdapter {
			public void mouseDragged(MouseEvent e) {
				mouseX = e.getX(); // track piece movement in mouseX, mouseY fields
				mouseY = e.getY();
				repaint(); // force repaint to animate motion
			}
		}

		// a keypress listener that listens for "u" undo commands
		class UndoHandler extends KeyAdapter {
			public void keyTyped(KeyEvent e) {
				if (reportMoves)
					System.out.println("undo");
				if (e.getKeyChar() == 'u') {
					if (stateStack.isEmpty()) 
						JOptionPane.showMessageDialog(null,"Cannot undo.");
					else {
						state = stateStack.pop();
						repaint();
					}
				}
			}
		}

	}
